/*
    list.h - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl> 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* A generic list type. In C++, this would be much neater. All elements must
   be of the same size (solve it with pointers, if needed) */

#ifndef PSICONV_LIST_H
#define PSICONV_LIST_H

#include <stddef.h>
#include <stdio.h>

/* Always use psiconv_list, never struct psiconv_list */
/* No need to export the actual internal format */
typedef struct psiconv_list *psiconv_list; 

/* Before using a list, call list_new. It takes the size of a single element
   as its argument. Always compute it with a sizeof() expression, just to be
   safe. The returned list is empty. */
extern psiconv_list psiconv_list_new(int element_size);

/* This frees the list. If elements contain pointers that need to be freed
   separately, call list_free_el below. */
extern void psiconv_list_free(psiconv_list l);

/* This calls free_el first for each element, before doing a list_free.
   Note that you should *not* do 'free(el)' at any time; that is taken care of
   automatically. */
extern void psiconv_list_free_el(psiconv_list l, void free_el(void *el));

/* Return the number of allocated elements */
extern int psiconv_list_length(const psiconv_list l);

/* Return 1 if the list is empty, 0 if not */
extern int psiconv_list_is_empty(const psiconv_list l);

/* Get an element from the list, and return a pointer to it. Note: you can
   directly modify this element, but be careful not to write beyond the
   element memory space. */
extern void * psiconv_list_get(const psiconv_list l, unsigned int indx);

/* Add an element at the end of the list. The element is copied from the
   supplied element. Of course, this does not help if the element contains
   pointers. */
extern void psiconv_list_add(psiconv_list l, void *el);

/* Do some action for each element. Note: you can directly modify the
   elements supplied to action, and they will be changed in the list,
   but never try a free(el)! */
extern void psiconv_list_foreach_el(psiconv_list l, void action(void *el));

/* Clone the list, that is, copy it. If elements contain pointers, you
   should call the next routine. */
extern psiconv_list psiconv_list_clone(const psiconv_list l);

/* Clone the list. For each element, clone_el is called. The elements which
   are given to clone_el must be modified in place, as needed. */
extern psiconv_list psiconv_list_clone_el(const psiconv_list l,
                                          void clone_el(void *el));

/* Read upto size_t elements from file f, and put them at the end of list l.
   Returned is the actual number of elements added. This assumes the file
   layout and the memory layout of elements is the same. */
extern size_t psiconv_list_fread(psiconv_list l,size_t size, FILE *f);

#endif
