/*
    parse_driver.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"
#include <stdlib.h>

#include "parse.h"
#include "parse_routines.h"
#include "data.h"

psiconv_file_type_t psiconv_file_type(psiconv_buffer buf)
{
  psiconv_header_section header;
  psiconv_file_type_t res;

  psiconv_parse_header_section(buf,0,0,NULL,&header);
  res = header->file;
  psiconv_free_header_section(header);
  return res;
}

int psiconv_parse(const psiconv_buffer buf,psiconv_file *result)
{
  int res=0;
  int lev=0;
  int off=0;

  (*result) = malloc(sizeof(**result));

  (*result)->type = psiconv_file_type(buf);
  if ((*result)->type == psiconv_unknown_file) {
    psiconv_warn(lev+1,off,"Unknown file type: can't parse!");
    (*result)->file = NULL;
    res = -1;
  } else if ((*result)->type == psiconv_word_file)
    res = psiconv_parse_word_file(buf,lev+2,off,
                                  (psiconv_word_f *)(&((*result)->file)));
  else if ((*result)->type == psiconv_texted_file)
    res = psiconv_parse_texted_file(buf,lev+2,off,
                                  (psiconv_texted_f *)(&((*result)->file)));
  else {
    psiconv_warn(lev+1,off,"Can't parse this file yet!");
    (*result)->file = NULL;
  }
    res = -1;

  return res;
}

int psiconv_parse_texted_file(const psiconv_buffer buf,int lev, psiconv_u32 off,
                              psiconv_texted_f *result)
{
  int res=0;
  psiconv_header_section header;
  psiconv_section_table_section table;
  psiconv_application_id_section appl_id;
  char *temp_str;
  psiconv_character_layout base_char;
  psiconv_paragraph_layout base_para;
  psiconv_u32 page_sec = 0;
  psiconv_u32 texted_sec = 0;
  psiconv_u32 applid_sec = 0;
  psiconv_section_table_entry entry;
  int i;

  psiconv_progress(lev+1,off,"Going to read a texted file");
  *result = malloc(sizeof(**result));

  psiconv_progress(lev+2,off, "Going to read the header section");
  res |= psiconv_parse_header_section(buf,lev+2,off,NULL,&header);

  psiconv_progress(lev+2,header->section_table_offset,
                   "Going to read the section table section");
  res |= psiconv_parse_section_table_section(buf,lev+2,
                                             header->section_table_offset,
                                             NULL,&table);

  for (i = 0; i < psiconv_list_length(table); i ++) {
    psiconv_progress(lev+2,header->section_table_offset,
                     "Going to read entry %d",i);
    entry = psiconv_list_get(table,i);
    if (entry->id == PSICONV_ID_APPL_ID_SECTION) {
      applid_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Application ID section at %08x",applid_sec);
    } else if (entry->id == PSICONV_ID_PAGE_LAYOUT_SECTION) {
      page_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Page Layout section at %08x",page_sec);
    } else if (entry->id == PSICONV_ID_TEXTED) {
      texted_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the TextEd section at %08x",texted_sec);
    } else {
      psiconv_warn(lev+3,header->section_table_offset,
                   "Found unknown section in the Section Table");
      psiconv_debug(lev+3,header->section_table_offset,
                    "Section ID %08x, offset %08x",entry->id,entry->offset);
      res = -1;
    }
  }
      
  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Application ID section");
  if (! applid_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Application ID section not found in the section table");
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Application ID section at offset %08x",applid_sec);
    res |= psiconv_parse_application_id_section(buf,lev+2,applid_sec,NULL,
                                                &appl_id);
  }
  if ((appl_id->id != PSICONV_ID_TEXTED) ||
       strcmp(appl_id->name,"TextEd.app")) {
    psiconv_warn(lev+2,applid_sec,
                 "Application ID section contains unexpected data");
    psiconv_debug(lev+2,applid_sec,"ID: %08x expected, %08x found",
                  PSICONV_ID_TEXTED,appl_id->id);
    temp_str = psiconv_make_printable(appl_id->name);
    psiconv_debug(lev+2,applid_sec,"Name: `%s' expected, `%s' found",
                            PSICONV_ID_TEXTED,temp_str);
    free(temp_str);
  }

  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Page layout section");
  if (! page_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Page layout section not found in the section table");
    (*result)->page_sec = NULL;
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Page layout section at offset %08x",page_sec);
    res |= psiconv_parse_page_layout_section(buf,lev+2,page_sec,NULL,
                                             &(*result)->page_sec);
  }

  base_char = psiconv_basic_character_layout();
  base_para = psiconv_basic_paragraph_layout();
  
  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the TextEd section");
  if (! texted_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "TextEd section not found in the section table");
   (*result)->texted_sec = NULL;
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "TextEd section at offset %08x",texted_sec);
    res |= psiconv_parse_texted_section(buf,lev+2,texted_sec,NULL,
                                        &(*result)->texted_sec,
                                        base_char,base_para);
  }
  psiconv_free_character_layout(base_char);
  psiconv_free_paragraph_layout(base_para);
  
  psiconv_free_application_id_section(appl_id);
  psiconv_free_header_section(header);
  psiconv_free_section_table_section(table);

  psiconv_progress(lev+1,off,"End of word file");
  return res;
}

int psiconv_parse_word_file(const psiconv_buffer buf,int lev, psiconv_u32 off,
                            psiconv_word_f *result)
{
  int res=0;
  psiconv_header_section header;
  psiconv_section_table_section table;
  psiconv_application_id_section appl_id;
  char *temp_str;
  psiconv_u32 pwd_sec = 0;
  psiconv_u32 status_sec = 0;
  psiconv_u32 styles_sec = 0;
  psiconv_u32 page_sec = 0;
  psiconv_u32 text_sec = 0;
  psiconv_u32 layout_sec = 0;
  psiconv_u32 applid_sec = 0;
  psiconv_section_table_entry entry;
  int i;

  psiconv_progress(lev+1,off,"Going to read a word file");
  *result = malloc(sizeof(**result));

  psiconv_progress(lev+2,off, "Going to read the header section");
  res |= psiconv_parse_header_section(buf,lev+2,off,NULL,&header);

  psiconv_progress(lev+2,header->section_table_offset, 
                   "Going to read the section table section");
  res |= psiconv_parse_section_table_section(buf,lev+2,
                                             header->section_table_offset,
                                             NULL,&table);

  for (i = 0; i < psiconv_list_length(table); i ++) {
    psiconv_progress(lev+2,header->section_table_offset,
                     "Going to read entry %d",i);
    entry = psiconv_list_get(table,i);
    if (entry->id == PSICONV_ID_APPL_ID_SECTION) {
      applid_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Application ID section at %08x",applid_sec);
    } else if (entry->id == PSICONV_ID_PAGE_LAYOUT_SECTION) {
      page_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Page Layout section at %08x",page_sec);
    } else if (entry->id == PSICONV_ID_TEXT_SECTION) {
      text_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Text section at %08x",text_sec);
    } else if (entry->id == PSICONV_ID_PASSWORD_SECTION) {
      pwd_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Password section at %08x",pwd_sec);
      psiconv_warn(lev+3,header->section_table_offset,
                   "Password section found - can't read encrypted data");
      res = -1;
    } else if (entry->id == PSICONV_ID_WORD_STATUS_SECTION) {
      status_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Word Status section at %08x",status_sec);
    } else if (entry->id == PSICONV_ID_WORD_STYLES_SECTION) {
      styles_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Word Styles section at %08x",styles_sec);
    } else if (entry->id == PSICONV_ID_LAYOUT_SECTION) {
      layout_sec = entry->offset;
      psiconv_debug(lev+3,header->section_table_offset,
                    "Found the Layout section at %08x",layout_sec);
    } else {
      psiconv_warn(lev+3,header->section_table_offset,
                   "Found unknown section in the Section Table");
      psiconv_debug(lev+3,header->section_table_offset,
                    "Section ID %08x, offset %08x",entry->id,entry->offset);
      res = -1;
    }
  }
      

  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Status section");
  if (!status_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Status section not found in the section table");
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Status section at offset %08x",status_sec);
    res |= psiconv_parse_word_status_section(buf,lev+2,status_sec,NULL,
                                             &((*result)->status_sec));
  }

  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Application ID section");
  if (! applid_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Application ID section not found in the section table");
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Application ID section at offset %08x",applid_sec);
    res |= psiconv_parse_application_id_section(buf,lev+2,applid_sec,NULL,
                                                &appl_id);
  }
  if ((appl_id->id != PSICONV_ID_WORD) ||
       strcmp(appl_id->name,"Word.app")) {
    psiconv_warn(lev+2,applid_sec,
                 "Application ID section contains unexpected data");
    psiconv_debug(lev+2,applid_sec,"ID: %08x expected, %08x found",
                  PSICONV_ID_WORD,appl_id->id);
    temp_str = psiconv_make_printable(appl_id->name);
    psiconv_debug(lev+2,applid_sec,"Name: `%s' expected, `%s' found",
                            PSICONV_ID_WORD,temp_str);
    free(temp_str);
  }

  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Page layout section");
  if (! page_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Page layout section not found in the section table");
   (*result)->page_sec = NULL;
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Page layout section at offset %08x",page_sec);
    res |= psiconv_parse_page_layout_section(buf,lev+2,page_sec,NULL,
                                             &(*result)->page_sec);
  }

  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Word Style section");
  if (!styles_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Word styles section not found in the section table");
   (*result)->styles_sec = NULL;
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Word styles section at offset %08x",styles_sec);
    res |= psiconv_parse_word_styles_section(buf,lev+2,styles_sec,NULL,
                                             &(*result)->styles_sec);
  }
  
  psiconv_progress(lev+2,header->section_table_offset,
                   "Looking for the Text section");
  if (!text_sec) {
   psiconv_warn(lev+2,header->section_table_offset,
                "Text section not found in the section table");
    (*result)->paragraphs = NULL;
    res = -1;
  } else {
    psiconv_debug(lev+2,header->section_table_offset,
                  "Text section at offset %08x",text_sec);
    res |= psiconv_parse_text_section(buf,lev+2,text_sec,NULL,
                                      &(*result)->paragraphs);
  }

  if (((*result)->paragraphs) && ((*result)->styles_sec)) {
    psiconv_progress(lev+2,header->section_table_offset,
                     "Looking for the Layout section");
    if (!layout_sec) {
      psiconv_debug(lev+2,header->section_table_offset,
                    "Layout section not found in the section table");
      res = -1;
    } else {
      psiconv_debug(lev+2,header->section_table_offset,
                    "Layout section at offset %08x",layout_sec);
      res |= psiconv_parse_styled_layout_section(buf,lev+2,layout_sec,NULL,
                                                 (*result)->paragraphs,
                                                 (*result)->styles_sec);
    }
  } else 
    psiconv_debug(lev+2,header->section_table_offset,
                  "Skipping search for Layout section, as either the "
                  "text or the word styles section was not found");

  psiconv_free_application_id_section(appl_id);
  psiconv_free_header_section(header);
  psiconv_free_section_table_section(table);

  psiconv_progress(lev+1,off,"End of word file");
  return res;
}
