/*
    parse_layout.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"
#include <stdlib.h>
#include <math.h>

#include "data.h"
#include "parse_routines.h"

int psiconv_parse_color(const psiconv_buffer buf, int lev, psiconv_u32 off, 
                        int *length, psiconv_color *result)
{
  int res = 0;
  int len = 0;

  psiconv_progress(lev+1,off,"Going to parse color");
  *result = malloc(sizeof(**result));

  (*result)->red = psiconv_read_u8(buf,lev+2,off+len);
  (*result)->green = psiconv_read_u8(buf,lev+2,off+len+1);
  (*result)->blue = psiconv_read_u8(buf,lev+2,off+len+2);
  len += 3;

  psiconv_debug(lev+2,off,"Color: red %02x, green %02x, blue %02x",
                (*result)->red, (*result)->green, (*result)->blue);
  if (length)
    *length = len;

  psiconv_progress(lev+1,off+len-1,"End of color (total length: %08x)",len);
  return res;
}



int psiconv_parse_font(const psiconv_buffer buf, int lev, psiconv_u32 off, 
                       int *length, psiconv_font *result)
{
  int res = 0;
  int strlength,i;
  char *str_copy;
  int len;

  psiconv_progress(lev+1,off,"Going to parse font");
  *result = malloc(sizeof(**result));

  strlength = psiconv_read_u8(buf,lev+2,off);
  (*result)->name = malloc(strlength);
  for (i = 0; i < strlength-1; i++)
    (*result)->name[i] = psiconv_read_u8(buf,lev+2,off + 1 + i);
  (*result)->name[strlength-1] = 0;
  (*result)->screenfont = psiconv_read_u8(buf,lev+2,off + strlength);

  str_copy = psiconv_make_printable((*result)->name);
  psiconv_debug(lev+2,off+1,"Found font `%s', displayed with screen font %02x",
                    str_copy,(*result)->screenfont);
  free(str_copy);
  len = strlength + 1;
  if (length)
    *length = len;

  psiconv_progress(lev+1,off + len - 1,"End of font (total length: %08x)",len);
  return res;
}

int psiconv_parse_border(const psiconv_buffer buf,int lev,psiconv_u32 off,
                         int *length, psiconv_border *result)
{
  int res = 0;
  int len = 0;
  psiconv_u32 temp;
  int leng;

  psiconv_progress(lev+1,off,"Going to parse border data");
  *result = malloc(sizeof(**result));

  psiconv_progress(lev+2,off+len,"Going to read border kind");
  temp = psiconv_read_u8(buf,lev+2,off+len);
  if (temp == 0x00)
    (*result)->kind = psiconv_border_none;
  else if (temp == 0x01)
    (*result)->kind = psiconv_border_solid;
  else if (temp == 0x02)
    (*result)->kind = psiconv_border_double;
  else if (temp == 0x03)
    (*result)->kind = psiconv_border_dotted;
  else if (temp == 0x04)
    (*result)->kind = psiconv_border_striped;
  else if (temp == 0x05)
    (*result)->kind = psiconv_border_dotstripe;
  else if (temp == 0x06)
    (*result)->kind = psiconv_border_dotdotstripe;
  else {
    psiconv_warn(lev+2,off,"Unknown border kind (defaults to `none')");
    (*result)->kind = psiconv_border_none;
    res = -1;
  }
  psiconv_debug(lev+2,off+len,"Kind: %02x",temp);
  len ++;

  psiconv_progress(lev+2,off+len,"Going to read border thickness");
  (*result)->thickness = psiconv_read_size(buf,lev+2,off+len,&leng);
  if (((*result)->kind != psiconv_border_solid) &&
      ((*result)->kind != psiconv_border_double) &&
      ((*result)->thickness != 0.0) &&
      (fabs((*result)->thickness - 1/20) >= 1/1000)) {
    psiconv_warn(lev+2,off,
                 "Border thickness specified for unlikely border type");
    res = -1;
  }
  psiconv_debug(lev+2,off+len,"Thickness: %f",(*result)->thickness);
  len += leng;

  psiconv_progress(lev+2,off+len,"Going to read the border color");
  res |= psiconv_parse_color(buf,lev+2,off+len,&leng,&(*result)->color);
  len += leng;
 
  psiconv_progress(lev+2,off+len,"Going to read the final unknown byte "
                                 "(0x01 expected)");
  temp = psiconv_read_u8(buf,lev+2,off + len);
  if (temp != 0x01) {
    psiconv_warn(lev+2,off,"Unknown last byte in border specification");
    psiconv_debug(lev+2,off+len, "Last byte: read %02x, expected %02x",
                  temp,0x01);
    res = -1;
  }
  len ++;

  if (length)
    *length = len;
 
  psiconv_progress(lev+1,off + len - 1,
           "End of border (total length: %08x)",len);

  return res;
}

int psiconv_parse_bullet(const psiconv_buffer buf,int lev,psiconv_u32 off,
                         int *length, psiconv_bullet *result)
{
  int res = 0;
  int len = 0;
  int leng;
  int bullet_length;

  *result = malloc(sizeof(**result));
  (*result)->on = psiconv_bool_true;

  psiconv_progress(lev+1,off,"Going to parse bullet data");
  psiconv_progress(lev+2,off+len,"Going to read bullet length");
  bullet_length = psiconv_read_u8(buf,lev+2,off+len);
  psiconv_debug(lev+2,off+len,"Length: %02x",bullet_length);
  len ++;

  psiconv_progress(lev+2,off+len,"Going to read bullet font size");
  (*result)->font_size = psiconv_read_size(buf,lev+2,off+len, &leng);
  len +=leng;
  
  psiconv_progress(lev+2,off+len,"Going to read bullet character");
  (*result)->character = psiconv_read_u8(buf,lev+2,off+len);
  psiconv_debug(lev+2,off+len,"Character: %02x",(*result)->character);
  len ++;
  
  psiconv_progress(lev+2,off+len,"Going to read indent on/off");
  res |= psiconv_parse_bool(buf,lev+2,off+len,&leng,&(*result)->indent);
  psiconv_debug(lev+2,off+len,"Indent on: %02x",(*result)->indent);
  len += leng;

  psiconv_progress(lev+2,off+len,"Going to read bullet color");
  res |= psiconv_parse_color(buf,lev+2,off+len,&leng,&(*result)->color);
  len += leng;

  psiconv_progress(lev+2,off+len,"Going to read bullet font");
  res |= psiconv_parse_font(buf,lev+2,off+len,&leng,&(*result)->font);
  len += leng;

  if (len != bullet_length + 1) {
    psiconv_warn(lev+2,off,"Bullet data structure length mismatch");
    psiconv_debug(lev+2,off,"Length: specified %02x, found %02x",
                  bullet_length,len-1);
    res = -1;
  }

  psiconv_progress(lev+1,off + len - 1,
                   "End of bullet data (total length: %08x)",len);

  if (length)
    *length = len;
  return res;
}

int psiconv_parse_tab(const psiconv_buffer buf, int lev, psiconv_u32 off,
                      int *length, psiconv_tab *result)
{
  int res = 0;
  int len = 0;
  int leng;
  psiconv_u8 temp;

  psiconv_progress(lev+1,off,"Going to parse tab");
  *result = malloc(sizeof(**result));

  psiconv_progress(lev+2,off,"Going to read tab location");
  (*result)->location = psiconv_read_length(buf,lev+2,off+len,&leng);
  len += leng;

  psiconv_progress(lev+2,off+len,"Going to read the tab kind");
  temp = psiconv_read_u8(buf,lev+2,off+len);
  if (temp == 1)
    (*result)->kind = psiconv_tab_left;
  else if (temp == 2)
    (*result)->kind = psiconv_tab_centre;
  else if (temp == 3)
    (*result)->kind = psiconv_tab_right;
  else {
    res = -1;
    psiconv_warn(lev+2,off+len,"Unknown tab kind argument");
    psiconv_debug(lev+2,off+len,"Kind found: %02x (defaulted to left tab)",
                  temp);
    (*result)->kind = psiconv_tab_left;
  }
  psiconv_debug(lev+2,off+len,"Kind: %02x",temp);
  len ++;
  
  if (length)
    *length = len;

  psiconv_progress(lev+1,off+len-1,"End of tab (total length: %08x)",len);
  return res;
  
}

int psiconv_parse_paragraph_layout_list(const psiconv_buffer buf, int lev,
                                        psiconv_u32 off, int *length,
                                        psiconv_paragraph_layout result)
{
  int res=0;
  int len=0;
  int list_length,leng,nr;
  psiconv_u8 id;
  psiconv_u32 temp;
  psiconv_tab tab;

  psiconv_progress(lev+1,off,"Going to read paragraph layout list");

  psiconv_progress(lev+2,off,"Going to read the list length");
  list_length = psiconv_read_u32(buf,lev+2,off + len);
  psiconv_debug(lev+2,off,"Length in bytes: %08x",list_length);
  len += 4;

  nr = 0;
  while(len - 4 < list_length) {
    psiconv_progress(lev+2,off+len,"Going to read element %d",nr);
    psiconv_progress(lev+3,off+len,"Going to read the element id");
    id = psiconv_read_u8(buf,lev+2,off+len);
    psiconv_debug(lev+3,off+len,"Id: %02x",id);
    len ++;
    switch(id) {
      case 0x01:
        psiconv_progress(lev+3,off+len,"Going to read background color");
        psiconv_free_color(result->back_color);
        res |= psiconv_parse_color(buf,lev+3,off+len,&leng,
                                   &result->back_color);
        len += leng;
        break;
      case 0x02:
        psiconv_progress(lev+3,off+len ,"Going to read indent left");
        result->indent_left = psiconv_read_length(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x03:
        psiconv_progress(lev+3,off+len,"Going to read indent right");
        result->indent_right = psiconv_read_length(buf,lev+2,off+len,&leng);
        len += leng;
        break;
      case 0x04:
        psiconv_progress(lev+3,off+len,"Going to read indent left first line");
        result->indent_first = psiconv_read_length(buf,lev+2,off+len, &leng);
        len += leng;
        break;
      case 0x05:
        psiconv_progress(lev+3,off+len,"Going to read horizontal justify");
        temp = psiconv_read_u8(buf,lev+3,off+len);
        if (temp == 0x00)
          result->justify_hor = psiconv_justify_left;
        else if (temp == 0x01)
          result->justify_hor = psiconv_justify_centre;
        else if (temp == 0x02)
          result->justify_hor = psiconv_justify_right;
        else if (temp == 0x03)
          result->justify_hor = psiconv_justify_full;
        else {
          res = -1;
          psiconv_warn(lev+3,off+len, "Unknown horizontal justify argument "
                       "in paragraph layout codes list");
        }
        psiconv_debug(lev+3,off+len,"Justify: %02x",temp);
        len ++;
        break;
      case 0x06:
        psiconv_progress(lev+3,off+len,"Going to read vertical justify");
        temp = psiconv_read_u8(buf,lev+3,off+len);
        if (temp == 0x00)
          result->justify_ver = psiconv_justify_top;
        else if (temp == 0x01)
          result->justify_ver = psiconv_justify_middle;
        else if (temp == 0x02)
          result->justify_ver = psiconv_justify_bottom;
        else {
          res = -1;
          psiconv_warn(lev+3,off+len, "Unknown vertical justify argument "
                                      "in paragraph layout codes list");
        }
        psiconv_debug(lev+3,off+len,"Justify: %02x",temp);
        len ++;
      case 0x07:
        psiconv_progress(lev+3,off+len,"Going to read interline distance");
        result->interline = psiconv_read_size(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x08:
        psiconv_progress(lev+3,off+len,"Going to read interline exact");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,
                                  &result->interline_exact);
        len += leng;
        break;
      case 0x09:
        psiconv_progress(lev+3,off+len,"Going to read top space");
        result->top_space = psiconv_read_size(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x0a:
        psiconv_progress(lev+3,off+len,"Going to read bottom space");
        result->bottom_space = psiconv_read_size(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x0b:
        psiconv_progress(lev+3,off+len,"Going to read on one page");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,
                                  &result->on_one_page);
        len += leng;
        break;
      case 0x0c:
        psiconv_progress(lev+3,off+len,"Going to read together with");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,
                                  &result->together_with);
        len += leng;
        break;
      case 0x0d:
        psiconv_progress(lev+3,off+len,"Going to read on next page");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,
                                  &result->on_next_page);
        len += leng;
        break;
      case 0x0e:
        psiconv_progress(lev+3,off+len,"Going to read no widow protection");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,
                                  &result->no_widow_protection);
        len += leng;
        break;
      case 0x10:
        psiconv_progress(lev+3,off+len,"Going to read border distance to text");
        result->border_distance = psiconv_read_length(buf,lev+3,
                                                         off+len,&leng);
        len += leng;
        break;
      case 0x11:
        psiconv_progress(lev+3,off+len,"Going to read top border");
        psiconv_free_border(result->top_border);
        res |= psiconv_parse_border(buf,lev+3,off+len,&leng,
                                    &result->top_border);
        len += leng;
        break;
      case 0x12:
        psiconv_progress(lev+3,off+len,"Going to read bottom border");
        psiconv_free_border(result->bottom_border);
        res |= psiconv_parse_border(buf,lev+3,off+len,&leng,
                                    &result->bottom_border);
        len += leng;
        break;
      case 0x13:
        psiconv_progress(lev+3,off+len,"Going to read left border");
        psiconv_free_border(result->left_border);
        res |= psiconv_parse_border(buf,lev+3,off+len,&leng,
                                    &result->left_border);
        len += leng;
        break;
      case 0x14:
        psiconv_progress(lev+3,off+len,"Going to read right border");
        psiconv_free_border(result->right_border);
        res |= psiconv_parse_border(buf,lev+3,off+len,&leng,
                                    &result->right_border);
        len += leng;
        break;
      case 0x15:
        psiconv_progress(lev+3,off+len,"Going to read bullet");
        psiconv_free_bullet(result->bullet);
        res |= psiconv_parse_bullet(buf,lev+3,off+len,&leng,
                                    &result->bullet);
        len += leng;
        break;
      case 0x16:
        psiconv_progress(lev+3,off+len,"Going to read standard tabs");
        result->tabs->normal = psiconv_read_length(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x17:
        psiconv_progress(lev+3,off+len,"Going to read extra tab");
        res |= psiconv_parse_tab(buf,lev+3,off+len,&leng,&tab);
        psiconv_list_add(result->tabs->extras,tab);
        len += leng;
        break;
      default:
        psiconv_warn(lev+3,off+len,
                     "Unknown code in paragraph layout codes list");
        psiconv_debug(lev+3,off+len,"Code: %02x",id);
        len ++;
        res = -1;
        break;
    }
    nr ++;
  }

  if (len - 4 != list_length) {
    psiconv_warn(lev+2,off+len,
         "Read past end of paragraph layout codes list. I probably lost track"
         "somewhere!");
    psiconv_debug(lev+2,off+len,"Read %d characters instead of %d",
                  len-4,list_length);
  }

  len = list_length + 4;

  psiconv_progress(lev+1,off+len,
                   "End of paragraph layout list (total length: %08x)",len);

  if (length)
    *length = len;
  return res;
}

int psiconv_parse_character_layout_list(const psiconv_buffer buf, int lev,
                                        psiconv_u32 off, int *length, 
                                        psiconv_character_layout result)
{
  int res=0;
  int len=0;
  int list_length,leng,nr;
  psiconv_u8 id;
  psiconv_u32 temp;

  psiconv_progress(lev+1,off,"Going to read character layout codes");

  psiconv_progress(lev+2,off,"Going to read the list length");
  list_length = psiconv_read_u32(buf,lev+2,off + len);
  psiconv_debug(lev+2,off,"Length in bytes: %08x",list_length);
  len += 4;

  nr = 0;
  while(len-4 < list_length) {
    psiconv_progress(lev+2,off+len,"Going to read element %d",nr);
    psiconv_progress(lev+3,off+len,"Going to read the element id");
    id = psiconv_read_u8(buf,lev+2,off+len);
    psiconv_debug(lev+3,off+len,"Id: %02x",id);
    len ++;
    switch(id) {
      case 0x19:
        psiconv_progress(lev+3,off+len,"Going to read text color");
        psiconv_free_color(result->color);
        res |= psiconv_parse_color(buf,lev+3,off+len, &leng,&result->color);
        len += leng;
        break;
      case 0x1a:
        psiconv_progress(lev+3,off+len,"Going to read background color (?)");
        psiconv_free_color(result->back_color);
        res |= psiconv_parse_color(buf,lev+2,off+len, &leng, 
                                   &result->back_color);
        len += leng;
        break;
      case 0x1c:
        psiconv_progress(lev+3,off+len,"Going to read font size");
        result->font_size = psiconv_read_size(buf,lev+3,off+len,&leng);
        len += leng;
        break;
      case 0x1d:
        psiconv_progress(lev+3,off+len,"Going to read italic");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,&result->italic);
        len += leng;
        break;
      case 0x1e:
        psiconv_progress(lev+3,off+len,"Going to read bold");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,&result->bold);
        len += leng;
        break;
      case 0x1f:
        psiconv_progress(lev+3,off+len,"Going to read super_sub");
        temp = psiconv_read_u8(buf,lev+3,off+len);
        if (temp == 0x00)
          result->super_sub = psiconv_normalscript;
        else if (temp == 0x01)
          result->super_sub = psiconv_superscript;
        else if (temp == 0x02)
          result->super_sub = psiconv_subscript;
        else {
          psiconv_warn(lev+3,off+len,
               "Unknown super_sub argument in character layout codes list");
          res = -1;
        }
        psiconv_debug(lev+3,off+len,"Super_sub: %02x",temp);
        len ++;
        break;
      case 0x20:
        psiconv_progress(lev+3,off+len,"Going to read underline");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,&result->underline);
        len += leng;
        break;
      case 0x21:
        psiconv_progress(lev+3,off+len,"Going to read strike_out");
        res |= psiconv_parse_bool(buf,lev+3,off+len,&leng,&result->strike_out);
        len += leng;
        break;
      case 0x22:
        psiconv_progress(lev+3,off+len,"Going to read font");
        psiconv_free_font(result->font);
        res |= psiconv_parse_font(buf,lev+3,off+len, &leng, &result->font);
        len += leng;
        break;
      default:
        psiconv_warn(lev+3,off+len,"Unknown code in character layout list");
        psiconv_debug(lev+3,off+len,"Code: %02x",id);
        len ++;
        res = -1;
        break;
    }
    nr ++;
  }

  if (len - 4 != list_length) {
    psiconv_warn(lev+2,off+len,
         "Read past end of character layout codes list. I probably lost track"
         "somewhere!");
    psiconv_debug(lev+2,off+len,"Read %d characters instead of %d",
                  len-4,list_length);
  }

  len = list_length + 4;

  psiconv_progress(lev+1,off+len,
                   "End of character layout list (total length: %08x)",len);

  if (length)
    *length = len;
  return res;
}
