/*
    parse_routines.h - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef PSICONV_PARSE_ROUTINES_H
#define PSICONV_PARSE_ROUTINES_H

#include "data.h"
#include "parse.h"

/* ***************
   * parse_aux.c *
   *************** */

/* These functions print error, warning, progress and debug information to
   stderr */
extern void psiconv_fatal(int level, psiconv_u32 off, const char *format,...);
extern void psiconv_warn(int level, psiconv_u32 off, const char *format,...);
extern void psiconv_progress(int level, psiconv_u32 off, 
                             const char *format,...);
extern void psiconv_debug(int level, psiconv_u32 off, const char *format,...);

/* This function returns a copy of a string, that contains no ASCII codes
   that are not printable. You should free this string yourself when you are
   done with it. */
extern char *psiconv_make_printable(const char *s);

/* **************
   * checkuid.c *
   ************** */

extern psiconv_u32 psiconv_checkuid(psiconv_u32 uid1,
                                    psiconv_u32 uid2,psiconv_u32 uid3);

/* ******************
   * parse_simple.c *
   ****************** */

extern psiconv_u8 psiconv_read_u8(const psiconv_buffer buf,int lev,
                                  psiconv_u32 off);
extern psiconv_u16 psiconv_read_u16(const psiconv_buffer buf,int lev,
                                    psiconv_u32 off);
extern psiconv_u32 psiconv_read_u32(const psiconv_buffer buf,int lev,
                                    psiconv_u32 off);

extern psiconv_S_t psiconv_read_S(const psiconv_buffer buf, int lev, 
                                  psiconv_u32 off, int *length);
extern psiconv_X_t psiconv_read_X(const psiconv_buffer buf, int lev, 
                                  psiconv_u32 off, int *length);
extern psiconv_length_t psiconv_read_length(const psiconv_buffer buf, int lev,
                                            psiconv_u32 off, int *length);
extern psiconv_size_t psiconv_read_size (const psiconv_buffer buf, int lev,
                                         psiconv_u32 off, int *length);
extern psiconv_string_t psiconv_read_string(const psiconv_buffer buf,int lev,
                                            psiconv_u32 off,int *length);

extern int psiconv_parse_bool(const psiconv_buffer buf, int lev, 
                              psiconv_u32 off, int *length, 
                              psiconv_bool_t *result);


/* ******************
   * parse_layout.c *
   ****************** */

extern int psiconv_parse_color(const psiconv_buffer buf, int lev, 
                               psiconv_u32 off, int *length, 
                               psiconv_color *result);

extern int psiconv_parse_font(const psiconv_buffer buf, int lev, 
                              psiconv_u32 off, int *length, 
                              psiconv_font *result);

extern int psiconv_parse_border(const psiconv_buffer buf,int lev,
                                psiconv_u32 off, int *length, 
                                psiconv_border *result);

extern int psiconv_parse_bullet(const psiconv_buffer buf,int lev,
                               psiconv_u32 off, int *length, 
                               psiconv_bullet *result);

extern int psiconv_parse_tab(const psiconv_buffer buf, int lev, 
                             psiconv_u32 off, int *length, 
                             psiconv_tab *result);

extern int psiconv_parse_paragraph_layout_list(const psiconv_buffer buf, 
                                   int lev, psiconv_u32 off, int *length,
                                   psiconv_paragraph_layout result);

extern int psiconv_parse_character_layout_list(const psiconv_buffer buf, 
                                   int lev, psiconv_u32 off, int *length,
                                   psiconv_character_layout result);


/* ****************
   * parse_page.c *
   **************** */

extern int psiconv_parse_page_header(const psiconv_buffer buf,int lev,
                                     psiconv_u32 off, int *length,
                                     psiconv_page_header *result);

extern int psiconv_parse_page_layout_section(const psiconv_buffer buf,int lev,
                                      psiconv_u32 off, int *length,
                                      psiconv_page_layout_section *result);

/* ******************
   * parse_common.c *
   ****************** */

extern int psiconv_parse_header_section(const psiconv_buffer buf,int lev,
                                        psiconv_u32 off, int *length, 
                                        psiconv_header_section *result);

extern int psiconv_parse_section_table_section(const psiconv_buffer buf, 
                                      int lev, psiconv_u32 off, 
                                      int *length,
                                      psiconv_section_table_section *result);

extern int psiconv_parse_application_id_section(const psiconv_buffer buf, 
                                      int lev, psiconv_u32 off, int *length,
                                      psiconv_application_id_section *result);

extern int psiconv_parse_text_section(const psiconv_buffer buf,int lev,
                                      psiconv_u32 off, int *length,
                                      psiconv_text_and_layout *result);

extern int psiconv_parse_styled_layout_section(const psiconv_buffer buf,
                                        int lev,psiconv_u32 off,
                                        int *length,
                                        psiconv_text_and_layout result,
                                        psiconv_word_styles_section styles);
extern int psiconv_parse_styleless_layout_section(const psiconv_buffer buf,
                                           int lev,psiconv_u32 off,
                                           int *length,
                                           psiconv_text_and_layout result,
                                           psiconv_character_layout base_char,
                                           psiconv_paragraph_layout base_para);


/* ******************
   * parse_texted.c *
   ****************** */

extern int psiconv_parse_texted_section(const psiconv_buffer buf,int lev,
                                        psiconv_u32 off, int *length,
                                        psiconv_texted_section *result,
                                        psiconv_character_layout base_char,
                                        psiconv_paragraph_layout base_para);


/* ****************
   * parse_word.c *
   **************** */

extern int psiconv_parse_word_status_section(const psiconv_buffer buf, int lev,
                                       psiconv_u32 off, int *length, 
                                       psiconv_word_status_section *result);

extern int psiconv_parse_word_styles_section(const psiconv_buffer buf, int lev,
                                       psiconv_u32 off, int *length,
                                       psiconv_word_styles_section *result);

/* ****************
   * parse_driver.c *
   **************** */

extern int psiconv_parse_texted_file(const psiconv_buffer buf,int lev, 
                                     psiconv_u32 off, psiconv_texted_f *result);

extern int psiconv_parse_word_file(const psiconv_buffer buf,int lev, 
                                   psiconv_u32 off, psiconv_word_f *result);


#endif /* PSICONV_PARSE_ROUTINES_H */
